/**
*  @license
* Copyright 2021, JsData. All rights reserved.
*
* This source code is licensed under the MIT license found in the
* LICENSE file in the root directory of this source tree.

* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* ==========================================================================
*/
import { alea } from 'seedrandom';
/**
 * Creates a new random number generator, optionally using the given seed.
 *
 * @param seed (Optional) the seed to be used.
 * @returns A new RNG method which returns floats from range `[0,1)`.
 */
export var createRng = function (seed) {
    return alea(seed === null || seed === void 0 ? void 0 : seed.toString());
};
/**
 * Take a uniform [0,1) random number generator (RNG) function and turns it into an
 * integer RNG function.
 *
 * @param rand An RNG in the range of [0,1).
 * @returns An RNG that takes an integer range (from, until) as arguments and returns
 *          a random integer in the range [from, until).
 */
export var randInt = function (rand) { return function (from, until) {
    if (from >= until)
        throw new Error('randInt(rng)(from,until): from must be less than until.');
    var result = (from + (until - from) * rand()) | 0;
    return result - +(result == until);
}; };
/**
 * Generates a random latin hypercube sampling, i.e. unique samples whose
 * features are evenly distributed over the range `[0,1)`. Latin hypercube
 * sampling is useful to create random non-duplicate samples that are
 * somewhat evenly distribute in the `[0,1)**nFeatures` hypercube.
 *
 * @see {@link https://en.wikipedia.org/wiki/Latin_hypercube_sampling}
 */
export var lhs = function (rand) { return function (nSamples, nFeatures) {
    var _randInt = randInt(rand);
    var result = [];
    // init columns to linspace(0,1, nSamples)
    for (var i = nSamples; i-- > 0;) {
        var row = new Float64Array(nFeatures);
        row.fill(i / (nSamples - 1 || 1));
        result.push(row);
    }
    // shuffle columns independently
    for (var n = void 0, i = nSamples; (n = i--) > 1;) {
        var ri = result[i];
        for (var j = nFeatures; j-- > 0;) {
            var k = _randInt(0, n), rij = ri[j];
            ri[j] = result[k][j];
            result[k][j] = rij;
        }
    }
    return result;
}; };
export var shuffle = function (rand) {
    return function (array) {
        var _randInt = randInt(rand);
        for (var i = array.length; i > 1;) {
            var j = _randInt(0, i--), tmp = array[i];
            array[i] = array[j];
            array[j] = tmp;
        }
    };
};
