/**
*  @license
* Copyright 2021, JsData. All rights reserved.
*
* This source code is licensed under the MIT license found in the
* LICENSE file in the root directory of this source tree.

* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* ==========================================================================
*/
import { convertToNumericTensor2D } from '../utils.js';
import { isScikit2D, assert, isDataFrameInterface } from '../typesUtils.js';
import { tensorMean, tensorStd, turnZerosToOnes } from '../math.js';
import { TransformerMixin } from '../mixins.js';
import { tf } from '../shared-esm/globals.js';
/**
 * Standardize features by removing the mean and scaling to unit variance.
 * The standard score of a sample x is calculated as: $z = (x - u) / s$,
 * where $u$ is the mean of the training samples, and $s$ is the standard deviation of the training samples.
 *
 * @example
 * ```js
 * import { StandardScaler } from 'scikitjs'
 *
 * const data = [
      [0, 0],
      [0, 0],
      [1, 1],
      [1, 1]
    ]

    const scaler = new StandardScaler()
    const expected = scaler.fitTransform(data)
    // const expected = [
    //  [-1, -1],
    //  [-1, -1],
    //  [1, 1],
    //  [1, 1]
    // ]
 * ```
 */
export class StandardScaler extends TransformerMixin {
    /** The per-feature scale that we see in the dataset. We divide by this number. */
    scale;
    /** The per-feature mean that we see in the dataset. We subtract by this number. */
    mean;
    /** Whether or not we should subtract the mean */
    withMean;
    /** Whether or not we should divide by the standard deviation */
    withStd;
    /** The number of features seen during fit */
    nFeaturesIn;
    /** The number of samples processed by the Estimator. Will be reset on new calls to fit */
    nSamplesSeen;
    /** Names of features seen during fit. Only stores feature names if input is a DataFrame */
    featureNamesIn;
    /** Useful for pipelines and column transformers to have a default name for transforms */
    name = 'standardscaler';
    constructor({ withMean = true, withStd = true } = {}) {
        super();
        this.withMean = withMean;
        this.withStd = withStd;
        this.scale = tf.tensor1d([]);
        this.mean = tf.tensor1d([]);
        this.nFeaturesIn = 0;
        this.nSamplesSeen = 0;
        this.featureNamesIn = [];
    }
    /**
     * Fit a StandardScaler to the data.
     */
    fit(X) {
        assert(isScikit2D(X), 'Data can not be converted to a 2D matrix.');
        const tensorArray = convertToNumericTensor2D(X);
        if (this.withMean) {
            this.mean = tensorMean(tensorArray, 0, true);
        }
        if (this.withStd) {
            const std = tensorStd(tensorArray, 0, true);
            // Deal with zero variance issues
            this.scale = turnZerosToOnes(std);
        }
        this.nSamplesSeen = tensorArray.shape[0];
        this.nFeaturesIn = tensorArray.shape[1];
        if (isDataFrameInterface(X)) {
            this.featureNamesIn = [...X.columns];
        }
        return this;
    }
    /**
     * Transform the data using the fitted scaler
     */
    transform(X) {
        assert(isScikit2D(X), 'Data can not be converted to a 2D matrix.');
        let tensorArray = convertToNumericTensor2D(X);
        if (this.withMean) {
            tensorArray = tensorArray.sub(this.mean);
        }
        if (this.withStd) {
            tensorArray = tensorArray.div(this.scale);
        }
        return tensorArray;
    }
    /**
     * Inverse transform the data using the fitted scaler
     */
    inverseTransform(X) {
        assert(isScikit2D(X), 'Data can not be converted to a 2D matrix.');
        let tensorArray = convertToNumericTensor2D(X);
        if (this.withStd) {
            tensorArray = tensorArray.mul(this.scale);
        }
        if (this.withMean) {
            tensorArray = tensorArray.add(this.mean);
        }
        return tensorArray;
    }
}
