"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeLowRankMatrix = exports.makeRegression = void 0;
const globals_1 = require("../shared-node/globals.js");
const makeRegression = ({ nSamples = 100, nFeatures = 100, nInformative = 10, nTargets = 1, noise = 1, bias = 0, effectiveRank = null, tailStrength = 0.5, shuffle = false, coef = false } = {}) => {
    return globals_1.tf.tidy(() => {
        const numberInformative = Math.min(nFeatures, nInformative);
        let X;
        if (effectiveRank === null) {
            // Randomly generate a well conditioned input set
            X = globals_1.tf.randomNormal([nSamples, nFeatures]);
        }
        else {
            X = (0, exports.makeLowRankMatrix)({
                nSamples,
                nFeatures,
                effectiveRank,
                tailStrength
            });
        }
        // Generate a ground truth model with only n_informative features being non
        // zeros (the other features are not correlated to y and should be ignored
        // by a sparsifying regularizers such as L1 or elastic net)
        const model = globals_1.tf.randomNormal([numberInformative, nTargets]).mul(100);
        const zeros = globals_1.tf.zeros([nFeatures - numberInformative, nTargets]);
        const groundTruth = globals_1.tf.concat([model, zeros]);
        let Y = X.dot(groundTruth).add(bias);
        // Add noise
        if (noise > 0) {
            Y = Y.add(globals_1.tf.randomNormal(Y.shape, undefined, noise));
        }
        // Randomly permute samples and features
        if (shuffle) {
            const randomTen = globals_1.tf.util.createShuffledIndices(nSamples);
            X = X.gather(randomTen);
        }
        Y = globals_1.tf.squeeze(Y);
        if (coef) {
            return [X, Y, globals_1.tf.squeeze(groundTruth)];
        }
        return [X, Y];
    });
};
exports.makeRegression = makeRegression;
const makeLowRankMatrix = ({ nSamples = 100, nFeatures = 100, effectiveRank = 10, tailStrength = 0.5 } = {}) => {
    return globals_1.tf.tidy(() => {
        let n = Math.min(nSamples, nFeatures);
        // Random (ortho normal) vectors
        let [u] = globals_1.tf.linalg.qr(globals_1.tf.randomNormal([nSamples, n]));
        let [v] = globals_1.tf.linalg.qr(globals_1.tf.randomNormal([nFeatures, n]));
        // Index of the singular values
        let singularIndex = globals_1.tf.range(0, n);
        // Build the singular profile by assembling signal and noise components
        const singularIndexByRank = singularIndex.div(effectiveRank);
        const lowRank = globals_1.tf
            .exp(singularIndexByRank.square().neg())
            .mul(1 - tailStrength);
        let tail = globals_1.tf.exp(singularIndexByRank.mul(-0.1)).mul(tailStrength);
        let s = lowRank.add(tail);
        return u.mul(s).dot(v.transpose());
    });
};
exports.makeLowRankMatrix = makeLowRankMatrix;
