"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BaseNaiveBayes = void 0;
/**
*  @license
* Copyright 2022, JsData. All rights reserved.
*
* This source code is licensed under the MIT license found in the
* LICENSE file in the root directory of this source tree.

* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* ==========================================================================
*/
const tfUtils_1 = require("../tfUtils.js");
const globals_1 = require("../shared-node/globals.js");
const utils_1 = require("../utils.js");
const serialize_1 = __importDefault(require("../serialize.js"));
class BaseNaiveBayes extends serialize_1.default {
    priors;
    varSmoothing;
    classes;
    means;
    variances;
    constructor(params = {}) {
        super();
        this.classes = globals_1.tf.tensor1d([]);
        this.means = [];
        this.variances = [];
        if (params.priors) {
            this.priors = (0, utils_1.convertToTensor1D)(params.priors);
        }
        this.varSmoothing = params.varSmoothing ? params.varSmoothing : 1e-9;
    }
    /**
     * Train the model by calculating the mean and variance of sample distribution.
     * @param X
     * @param y
     * @returns
     */
    async fit(X, y) {
        const features = (0, utils_1.convertToNumericTensor2D)(X);
        const labels = (0, utils_1.convertToTensor1D)(y);
        const { values, meansByLabel, variancesByLabel } = globals_1.tf.tidy(() => {
            (0, tfUtils_1.polyfillUnique)(globals_1.tf);
            const meansByLabel = [];
            const variancesByLabel = [];
            // Get the list of unique labels
            const { values } = globals_1.tf.unique(labels);
            const { variance } = globals_1.tf.moments(features, 0);
            const epsilon = variance.max().mul(this.varSmoothing);
            globals_1.tf.unstack(values).forEach((c) => {
                const mask = globals_1.tf.equal(labels, c).toFloat();
                const numInstances = globals_1.tf.sum(mask);
                const mean = globals_1.tf
                    .mul(features, mask.expandDims(1))
                    .sum(0)
                    .div(numInstances);
                const variance = globals_1.tf
                    .sub(features, mean)
                    .mul(mask.expandDims(1))
                    .pow(2)
                    .sum(0)
                    .div(numInstances)
                    .add(epsilon);
                meansByLabel.push(mean);
                variancesByLabel.push(variance);
            });
            return { values, meansByLabel, variancesByLabel };
        });
        // Unique labels this model have learned
        this.classes = values;
        this.means = meansByLabel;
        this.variances = variancesByLabel;
        return this;
    }
    /**
     * Predict the probability of samples assigned to each observed label.
     * @param X
     * @returns {tf.Tensor} Probabilities
     */
    predictProba(X) {
        const features = (0, utils_1.convertToNumericTensor2D)(X);
        const probabilities = globals_1.tf.tidy(() => {
            let probs = [];
            this.classes.unstack().forEach((_, idx) => {
                // Get the mean for this label
                const mean = this.means[idx];
                const variance = this.variances[idx];
                const prob = this.kernel(features, mean, variance);
                probs.push(prob);
            });
            const withoutPriors = globals_1.tf.stack(probs, 1);
            if (this.priors) {
                return withoutPriors.mul(this.priors);
            }
            else {
                return withoutPriors;
            }
        });
        return probabilities;
    }
    /**
     * Predict the labels assigned to each sample
     * @param X
     * @returns {tf.Tensor} Labels
     */
    predict(X) {
        const probs = this.predictProba(X);
        return probs.argMax(1);
    }
    toJson() {
        const jsonClass = JSON.parse(super.toJson());
        if (this.priors) {
            jsonClass.priors = this.priors.arraySync();
        }
        jsonClass.classes = this.classes.arraySync();
        jsonClass.means = this.means.map((t) => t.arraySync());
        jsonClass.variances = this.variances.map((v) => v.arraySync());
        return JSON.stringify(jsonClass);
    }
    fromJson(model) {
        const jsonModel = JSON.parse(model);
        if (jsonModel.priors) {
            jsonModel.priors = globals_1.tf.tensor(jsonModel.priors);
        }
        jsonModel.classes = globals_1.tf.tensor(jsonModel.classes);
        const means = [];
        for (const wMeans of jsonModel.means) {
            means.push(globals_1.tf.tensor(wMeans));
        }
        const variances = [];
        for (const variance of jsonModel.variances) {
            variances.push(globals_1.tf.tensor(variance));
        }
        jsonModel.means = means;
        jsonModel.variances = variances;
        return Object.assign(this, jsonModel);
    }
}
exports.BaseNaiveBayes = BaseNaiveBayes;
