"use strict";
/**
*  @license
* Copyright 2021, JsData. All rights reserved.
*
* This source code is licensed under the MIT license found in the
* LICENSE file in the root directory of this source tree.

* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* ==========================================================================
*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.BruteNeighborhood = void 0;
const globals_1 = require("../shared-node/globals.js");
const typesUtils_1 = require("../typesUtils.js");
/**
 * A {@link Neighborhood} implementation that uses a brute force approach
 * to nearest neighbor search. During a {@link BruteNeighborhood#kNearest}
 * query, the distance between every entry and the query point is computed.
 */
class BruteNeighborhood {
    _metric;
    _entries;
    constructor({ metric, entries }) {
        this._metric = metric;
        this._entries = entries;
    }
    kNearest(k, queryPoints) {
        const { _metric, _entries } = this;
        (0, typesUtils_1.assert)(_entries.shape[1] == queryPoints.shape[1], 'X_train.shape[1] must equal X_predict.shape[1]');
        // // batched version
        // const [m, n] = queryPoints.shape
        // return tf.tidy(() => {
        //   const negDist = _metric.tensorDistance(
        //     queryPoints.reshape([m, 1, n]),
        //     _entries
        //   ).neg() as Tensor2D
        //   const { values, indices } = tf.topk(negDist, k)
        //   return { distances: values.neg(), indices }
        // })
        // unbatched version
        return globals_1.tf.tidy(() => {
            const result = globals_1.tf.unstack(queryPoints).map((queryPoint) => {
                return globals_1.tf.tidy(() => {
                    const dist = _metric.tensorDistance(queryPoint, _entries).neg();
                    const { values, indices } = globals_1.tf.topk(dist, k);
                    return [values, indices];
                });
            });
            return {
                distances: globals_1.tf.stack(result.map((x) => x[0])).neg(),
                indices: globals_1.tf.stack(result.map((x) => x[1]))
            };
        });
    }
}
exports.BruteNeighborhood = BruteNeighborhood;
